;Beispiel4 zur Anwendung von AssembleIt2 im 64Bit-Modus
;Im 64Bit-Modus von Autoit kann nicht direkt auf den 32Bit-Assembler (in Form einer Dll) zugegriffen werden.
;Daher wird die Datei AssembleIt2_Helper64.exe (welche den Assembler enthält) gestartet und an diese Datei
;per Read/WriteProcessMemory() der Code übergeben, dort assembliert und wieder zurückgegeben.
;Es sind sowohl Macros, als auch Variablen benutzbar.
;ACHTUNG! Da im 64Bit-Modus FastCall verwendet wird, diese Aufrufkonvention unbedingt beachten
;Oder das unten vorgestellte Macro GetArgs64Fastcall einbinden^^
;FLOAT/DOUBLE werden in den Registern XMM0-XMM4 übergeben! Restliche Parameter über den Stack!
;
;Da im 64Bitmodus der Assembler aufwendig eingebunden werden muss und bei der Ausführung einige Millisekunden Overhead
;entstehen, empfiehlt sich die Entwicklung mit Hinblick auf Code, der per DllCallAddress() ausgeführt werden kann, um
;die volle Geschwindigkeit des 64Bit-Modus ausnutzen zu können!

;
;Für 32Bit-Modus mus gesetzt werden #AutoIt3Wrapper_UseX64=n
;Für 64Bit-Modus mus gesetzt werden #AutoIt3Wrapper_UseX64=y


#AutoIt3Wrapper_UseX64=y                  ;64Bit-Modus

#include-once
#include <assembleit2_64.au3>
#include <SliderConstants.au3>


#cs
    Beispielscript um eine mit ASM-Code erstellte Farbpalette in einer Bitmap darzustellen
    Innerhalb von 2 Schleifen werden die Grün- und Blauwerte von Pixeln gesetzt,
    der Rot-Anteil des Pixels wird durch einen Slider in der GUI festgelegt und an den Code übergeben.
    Weiterhin wird die Startadresse der Bitmap an den Code übergeben.

    Vorgestellt sind mehrere Varianten, um ASM-Code mittels Assembleit2 ablaufen zu lassen.

    1. Beispiel zur Verwendung von Macros und in ASM definierten Variablen (kein DllCallAddress!)
    2. Beispiel als auch per DllCallAddress ausführbaren Code (sehr schnell!)
    3. Beispiel mit Verwendung von AutoIt-Variablen im ASM-Code (kein DllCallAddress!)

    Der Overhead von mit AssembleIt2 laufenden Scripten beträgt nur eine Handvoll Millisekunden (2-3ms)!
    In dieser Zeit wird der ASM-Code zur Laufzeit assembliert, in den Speicher geschrieben und dort per DllCallAddress ausgeführt.

    AssembleIt2 ist auch in der Lage, den Code zu assemblieren und in einer Variable oder Datei abzuspeichern.
    Diesen Code kann man in einer Struct speichern um ihn per DllCallAddress() direkt im Speicher auszuführen.
    Dies ist die zzt. schnellste Variante, um Assemblerprogramme auszuführen. AssembleIt2 und sämtliche dafür
    notwendigen Includes sind nicht mehr nötig!

    Bei extrem aufwendigen Berechnungen bleibt daher abzuwägen, ob man die Vorteile der Verwendung von
    Macros und (AutoIt-)Variablen im Code der Erstellung von mit DllCallAddress()lauffähigen Code vorzieht!


    Vom assemblierten Code wird die Anzahl der Prozessortakte zurückgegeben. Bitte beachten, dass die Laufzeit des Codes
    so gering ist, dass moderne Prozessoren im Sleep-Modus bleiben und daher nur mit minimalem Takt laufen!

#ce


;#####################################################################################################
;Beispiel4 zur Verwendung von Macros und definierten ASM-Variablen, nur in Verbindung mit AssembleIt2.
;Code nicht ohne Änderungen mit DllCallAddress lauffähig!
;#####################################################################################################


#cs _64Bitmacros;                          ;funktioniert nur als "AssembleIt2"-Funktion, da "interne" Variablen verwendet werden

    use64                                 ;bei 64 Bit werden in rcx,rdx,r8,r9 die ersten 4 parameter übergeben

    macro GetArgs64Fastcall               ;ein Macro, um die Parameter zuzuweisen
    {
    mov [Arg1],rcx                        ;bei fastCall werden die ersten 4 Parameter in Registern übergeben...
    mov [Arg2],rdx
    mov [Arg3],r8
    mov [Arg4],r9
    mem2mem64 [Arg5],[rsp+40]             ;die anderen parameter über den Stack
    mem2mem64 [Arg6],[rsp+48]             ;Stack = 8 Byte rücksprungadresse+32 bytes Shadowspace
    mem2mem64 [Arg7],[rsp+56]             ;der 5. Parameter fängt ab [rsp+40] an
    mem2mem64 [Arg8],[rsp+64]
    }

    macro mem2mem64 mem1,mem2             ;Makro um einen Speicherinhalt in einen anderen zu "moven"
    {                                     ;die in den Macros verwendeten Registerinhalte sind natürlich nicht local
    mov rax,mem2                          ;der Inhalt von EAX ist auch ausserhalb der Macros gültig!
    mov mem1,rax
    }



    org $PTR_SOURCE_ASMCODE               ;Basisadresse auslesen, um die Variablenadressen zuzuweisen

    GetArgs64Fastcall                     ;[Arg1]=rcx  , [Arg2]=rdx ,[Arg3]=r8  uswusf

    push rdi                              ;nichtflüchtige register retten
    push rsi
    push rbx
    push rbp

    mov rdi,[Arg1]                        ;pointer pixel
    mov rcx,[Arg2]                        ;sliderwert, roter Farbanteil des Pixels

    rdtsc                                 ;Ticks1 = edx:eax um Takte zu zählen ^^
    push rax                              ;Ticks1 auf Stack speichern

    xor rax,rax                           ;rax=0
    align 16
    mov rdx,256                           ;x=0
    _xschleife:                           ;for x=0 to 255  grün

    mov rbx,256                           ;y=0
    _yschleife:                           ;for y=0 to 255   ;blau

    ;Pixelbytes AARRGGBB zusammensetzen

    mov ah,cl                             ;0000RR00  rot
    shl eax,8                             ;AARR0000  8 Bit nach links schieben
    mov ah,dl                             ;AARRGG00  grün
    mov al,bl                             ;AARRGGBB  blau
    stosd                                 ;ersetzt die nächsten beiden Zeilen, schreibt EAX an [EDI] und erhöht EDI um 4
    ;~     mov [rdi],rax                       ;pixel an adresse schreiben
    ;~     add rdi,4                           ;nächstes pixel

    sub rbx,1                             ;blau=blau-1 , sub setzt im Gegensatz zu dec die Flags,
    jnz _yschleife                        ;JumpNotZero nach label, das Zero-Flag wird hier abgefragt

    sub rdx,1                             ;grün=grün-1
    jnz _xschleife                        ;Grün<>0? ja, dann nächstes grün (Sprung zum Label)

    rdtsc                                 ;Ticks= edx:eax aktueller Zählerstand der Ticks seit Rechnerstart
    pop rbx                               ;Ticks wurden am Anfang des Scriptes auf den Stack gespeichert
    sub rax,rbx                           ;Ticks=Ticks-Ticks1 Anzahl benötigter Takte seit Start zurückgeben

    pop rbp
    pop rbx
    pop rsi
    pop rdi

    ret                                   ;Ende



    ;Variablen als QUADWORD

    Arg1 dq ?
    Arg2 dq ?
    Arg3 dq ?
    Arg4 dq ?
    Arg5 dq ?
    Arg6 dq ?
    Arg7 dq ?
    Arg8 dq ?

#ce



$gui = GUICreate("Beispiel", 296, 256)

$slider = GUICtrlCreateSlider(261, 5, 30, 251, BitOR($TBS_VERT, $TBS_TOP, $TBS_LEFT, $TBS_BOTH, $TBS_NOTICKS, $TBS_ENABLESELRANGE))
GUICtrlSetLimit(-1, 255, 0)

GUISetState()

$DC_gui = _WinAPI_GetDC($gui)             ;DeviceContext der GUI holen
Global $ptr, $hbmp                        ;pointer auf die pixel, handle bitmap
$DC_bitmap = _CreateNewBmp32(256, 256, $ptr, $hbmp) ;bitmap erstellen


While GUIGetMsg() <> -3

    $sliderpos = GUICtrlRead($slider)     ;sliderposition

    $t = TimerInit()
    Local $ret = _AssembleIt2("uint_ptr", "_64Bitmacros", "ptr", $ptr, "int_ptr", $sliderpos) ;, "int_ptr", 33, "int_ptr", 44, "int_ptr", 55, "int_ptr", 66) ;64Bit-Modus
    ConsoleWrite("_AssembleIt2()  64Bit      Anzahl Takte = " & ($ret) & "     Zeit [ms] = " & Int(1000 * TimerDiff($t)) / 1000 & @CRLF) ;### Debug Console

    _WinAPI_BitBlt($DC_gui, 0, 0, 256, 256, $DC_bitmap, 0, 0, 0xCC0020) ;$srccopy

WEnd

ProcessClose("AssembleIt2_Helper64.exe")
_DeleteBitmap32($DC_bitmap, $ptr, $hbmp)



Func _CreateNewBmp32($iwidth, $iheight, ByRef $ptr, ByRef $hbmp) ;erstellt leere 32-bit-Bitmap; Rückgabe $HDC und $ptr und handle auf die Bitmapdaten
    $hcdc = _WinAPI_CreateCompatibleDC(0) ;Desktop-Kompatiblen DeviceContext erstellen lassen
    $tBMI = DllStructCreate($tagBITMAPINFO) ;Struktur der Bitmapinfo erstellen und Daten eintragen
    DllStructSetData($tBMI, 1, DllStructGetSize($tBMI) - 4) ;Structgröße abzüglich der Daten für die Palette
    DllStructSetData($tBMI, 2, $iwidth)
    DllStructSetData($tBMI, 3, -$iheight) ;minus =standard = bottomup
    DllStructSetData($tBMI, 4, 1)
    DllStructSetData($tBMI, 5, 32)        ;32 Bit = 4 Bytes => AABBGGRR
    $adib = DllCall('gdi32.dll', 'ptr', 'CreateDIBSection', 'hwnd', 0, 'ptr', DllStructGetPtr($tBMI), 'uint', 0, 'ptr*', 0, 'ptr', 0, 'uint', 0)
    $hbmp = $adib[0]                      ;hbitmap handle auf die Bitmap, auch per GDI+ zu verwenden
    $ptr = $adib[4]                       ;pointer auf den Anfang der Bitmapdaten, vom Assembler verwendet
    _WinAPI_SelectObject($hcdc, $hbmp)    ;objekt hbitmap in DC
    Return $hcdc                          ;DC der Bitmap zurückgeben
EndFunc                                   ;==>_CreateNewBmp32

Func _DeleteBitmap32($DC, $ptr, $hbmp)
    _WinAPI_DeleteDC($DC)
    _WinAPI_DeleteObject($hbmp)
    $ptr = 0
EndFunc                                   ;==>_DeleteBitmap32




























